/* saladshooter.c - a word salad generator */
/* Created: Wed Dec 13 11:40:02 AM EST 2023 malakai */
/* Copyright © 1991-2023 The Last Outpost Project */
/* Based on salad.c from "Strange Attractor" Copyright © 2015
 * <rahjiii@jeffrika.com>, used with permission. */

/* Sun Dec 17 03:37:27 PM EST 2023
 * This code is taken from the Last Outpost Project and made available by the
 * author through the Last Outpost Code Archive: <rahjiii@jeffrika.com>
 * (https://www.last-outpost.com/LO/pubcode) Feel free to use it as you see
 * fit. */

/* $Id: saladshooter.c,v 1.1 2023/12/17 20:33:05 malakai Exp $*/

/* ---- standard include files ---- */
#include <stdio.h>
#include <json-c/json.h>
#include <ctype.h>

/* These includes are for integration with the game, and should define
 * everthing listed in the USE_LOCAL_DEFINITIONS sections, and everything
 * needed by the LO/DIKU example functions, if you are using those. */
#include "structs.h"
#include "db.h"
#include "utils.h"
#include "scnprintf.h"
#include "debug.h"
#include "comm.h"
#include "constants.h"

/* this is saladshooter's own header. */
#include "saladshooter.h"

/* ---- local #defines ---- */
/* a simple recursion-breaker */
#define MAX_RECURSION 100

/* saladshooter.c needs the following items to be defined somewhere, either
 * right here below this comment, or in a game specific include file, which may
 * be listed above. */
#if USE_LOCAL_DEFINITITIONS
#define lo_log(args...) printf(args)
#define scnprintf(args...) snprintf(args)
#define number(from,to) (rand()%((to)-(from)+1) + (from))
#endif /* USE_LOCAL_DEFINITIONS */

/* ---- structs and typedefs ---- */

/* For simplicity, saladshooter.c drops some of the functionality from the
 * original salad.c. Notably the random number generator seed argument which
 * was used for repeatable procedural generation, the merged multi-level
 * multi-database variable heirarchy, and one-time-only variable substituion.
 * The scoped variables from salad.c (like $title:corporate or
 * $federation:person) are NOT supported, and variables WILL be re-evaluated
 * each time they are seen instead of the first time they are seen.  In this
 * code, "$name is called $name" could resolve to "Bob is called Tom" instead
 * of "Bob is called Bob" each time.  */

/* The format of a saladshooter.c database is a 'flat' JSON array of variable
 * names and their substitutions.  In a substitution, A variable begins with a
 * '$' and ends with a space or a '+'. */

/* ---- local variable declarations ---- */
json_object *main_saladdb;

/* ---- local function declarations ---- */
void saladshooter_init(char *filename);
int saladshooter(char *out, size_t outlen, const char *from);
void saladshooter_free(void);
json_object *saladshooter_loaddb(char *filename);
int saladshooter_json(char *out, size_t outlen, json_object *dbobj, const char *from, int depth);

/* ---- code starts here ---- */

/* Load a default datatbse from the filename. */
void saladshooter_init(char *filename) {
	main_saladdb =  saladshooter_loaddb(filename);
	return;
}

/* This calls the saladshooter_json() function with the main_saladdb as the
 * database. This is so that callers don't need to specifiy the default db
 * every time, and to avoid the caller having to include json-c.  */
int saladshooter(char *out, size_t outlen, const char *from) {
	return(saladshooter_json(out,outlen,main_saladdb,from,0));
}

/* De-allocate the default database. */
void saladshooter_free(void) {
	if(main_saladdb) {
		json_object_put(main_saladdb);
		main_saladdb = NULL;
	}
}

/* loads a json object from a filename, with localized error printing. */
json_object *saladshooter_loaddb(char *filename) {

	json_object *jobj = json_object_new_object();

	if(!(jobj=json_object_from_file(filename))) {
		lo_log("saladshooter_loaddb() problem loading '%s': %s",filename,json_util_get_last_err());
		return(NULL);
	}
	return(jobj);
}


/* Copies from to out, recursively matching $ directives. The dbobj is a
 * json_object for the saladdb the caller wants to use.  See saladshooter_json()
 * for using the default db that was set up during init.*/
int saladshooter_json(char *out, size_t outlen, json_object *dbobj, const char *from, int depth) {

	char *eob;
	char *o,*s;
	const char *f;
	char *sub;
	const char *rstr;
	json_object *aobj;
	int idx;

	if(!out) {
		return(0);
	}
	*out='\0';
	if(!dbobj || !from || (outlen == 0)) {
		return(0);
	}

	if(depth>MAX_RECURSION) {
		lo_log("saladshooter() loop detected processing '%s'.",from);
		return(0);
	}

	eob = out + outlen -1;
	o = out;
	f = from;

	/* copy from to out unless you find a magic token $. */
	while( (o<=eob) && (*f) && (*f != '$') ) {
		*o++ = *f++;
	}

	if(*f=='\0' || o>=eob) {
		*o = '\0';
		/* done baby done! */
		return(o-out);
	}

	/* found a variabile substitution to make */
	if(*f=='$') {
		f++; /* skip over it */
		/* parse for the variable name. */
		sub = (char *)malloc(sizeof(char)*(outlen+1));
		s=sub;
		while( (*f) && (*f != '+') && (isalnum(*f) || *f=='_') ) {
			*s++ = *f++;
		}
		*s++ = '\0';

		/* + is a 'this ends the variable name without a space' character, so
		 * you can do things like "$goname+ian" */
		if(*f == '+') f++;
		/* now sub contains the variable name for the next string substitution. */
		/* do any special handling here... */

		/* now its just a random string sub. */
		if(!json_object_object_get_ex(dbobj,sub,&aobj)) {
			/* If there is no such variable, you could call it an error... */
			/*   lo_log("%s ERROR: no such variable '$%s' in '%s'",__func__,sub,from);   */
			/* or just leave it in the data stream.  Maybe its a color code or
			 * something. */
			o += scnprintf(o,eob-o,"$%s",sub);
		} else {
			idx = number(0,json_object_array_length(aobj)-1);
			rstr = json_object_get_string(json_object_array_get_idx(aobj,idx));
			o += saladshooter_json(o,eob-o,dbobj,rstr,depth+1);
		}
		free(sub);

		if(o>=eob) {
			return(outlen);
		}

	}
	if(*f=='\0') {
		return(o-out);
	}

	/* otherwise  process the remainder of the string. */
	o += saladshooter_json(o,eob-o,dbobj,f,depth+1);
	if(o>=eob) {
		return(outlen);
	}
	return(o-out);

}

/* a LO/DIKU in game test function.  */
void do_saladname(struct char_data * ch, char *argument, int cmd) {

	char buf[MAX_STRING_LENGTH];
	char *strp = (char *)malloc(sizeof(char) * strlen(argument));

	char *a,*b;
	a = argument;
	b = strp;
	while(*a) {
		*b++ = *a++;
		if(*(b-1) == '$' && *a=='$') {
			a++;
		}
	}
	*b='\0';

	act("$cg$cUSome names:$cn",FALSE,ch,NULL,NULL,TO_INFO);

	for(int i=0;i<10;i++) {
		*buf='\0';
		saladshooter(buf,sizeof(buf),strp);
		send_to_char(buf,ch);
		send_to_char("\r\n",ch);
	}
	act("Done.",FALSE,ch,NULL,NULL,TO_INFO);

	free(strp);
}

/* A LO/DIKU function to choose a name for a char based on race and sex */
int fakename_for_char(char *out, size_t outlen,int race,int sex) {

	char from[MAX_STRING_LENGTH];
	
	if(race > RACE_GNOME) race=RACE_ELF;
	if((sex == SEX_NEUTRAL) || (sex > SEX_PLURAL)) sex=SEX_PLURAL;

	/* prints something like "$Ogre_Female", which must exist in the db. */
	scnprintf(from,sizeof(from),"$%s_%s",races[race],sexes[sex]);

	return(saladshooter(out,outlen,from));

}

